"""
SAMPLE BEGINNING OF FILE DOCSTRING
"""

import torch
import torch.nn as nn
import numpy as np 
from sklearn.preprocessing import OneHotEncoder, StandardScaler
from sklearn.impute import SimpleImputer
from sklearn.compose import make_column_transformer
from sklearn.pipeline import make_pipeline
from tqdm import tqdm
from sksurv.nonparametric import kaplan_meier_estimator
from sklearn.model_selection import train_test_split
from dnamite.loss_fns import drsa_loss

class DRSASingleSplit(nn.Module):
    def __init__(self, n_input, n_hidden, n_output, eval_times, device="cpu"):
        super().__init__()
        self.n_input = n_input
        self.n_output = n_output
        self.n_hidden = n_hidden
        self.eval_times = eval_times
        
        self.rnn = nn.LSTM(n_input+1, n_hidden, batch_first=True)
        self.output_head = nn.Sequential(
            nn.Linear(n_hidden, n_hidden),
            nn.ReLU(),
            nn.Linear(n_hidden, 1)
        )
        
    def forward(self, x):
        # x is shape (N, p)
        
        # Add eval times to x
        inputs = x.unsqueeze(1).repeat(1, self.n_output, 1)
        inputs = torch.cat([
            inputs,
            self.eval_times.unsqueeze(0).repeat(x.shape[0], 1).unsqueeze(-1)
        ], dim=2)
        
        h, _ = self.rnn(inputs)
        
        return self.output_head(h)
    
    def train_(
        self, 
        train_epoch_fn, 
        test_epoch_fn,
        train_loader,
        test_loader,
        optimizer,
        n_epochs,
    ):
        
        early_stopping_counter = 0
        best_test_loss = float('inf')

        for epoch in range(n_epochs):
            train_epoch_fn(self, train_loader, optimizer)
            test_loss, test_preds = test_epoch_fn(self, test_loader)
            
            # print(f"Epoch {epoch+1} | Train loss: {train_loss:.3f} | Test loss: {test_loss:.3f} | Num Feats: {len([z for z in self.get_smooth_z() if z > 0])}")

            # Check if the test loss has improved
            if test_loss < best_test_loss:
                best_test_loss = test_loss
                early_stopping_counter = 0
                
                # Save the model at the best test loss
                torch.save(self.state_dict(), "../model_saves/tmp_best_model.pt")
                
            else:
                early_stopping_counter += 1

            # If test loss has not improved for 5 consecutive epochs, terminate training
            if early_stopping_counter >= 5:
                print(f"Early stopping at {epoch+1} epochs: Test loss has not improved for 5 consecutive epochs.")
                break
            
        # Load the model from the best test loss
        self.load_state_dict(torch.load("../model_saves/tmp_best_model.pt"))

        return


class DRSA(nn.Module):
    """  
    DRSA
    """
    
    def __init__(
        self, 
        n_input, 
        n_hidden, 
        n_output, 
        n_eval_times,
        validation_size=0.2,
        n_val_splits=5,
        learning_rate=1e-4,
        max_epochs=100,
        batch_size=128,
        device="cpu",
        **kwargs
    ):
        super().__init__()
        self.n_input = n_input
        self.n_hidden = n_hidden
        self.n_eval_times = n_eval_times
        self.validation_size = validation_size
        self.n_val_splits = n_val_splits
        self.learning_rate = learning_rate
        self.max_epochs = max_epochs
        self.batch_size = batch_size
        self.device = device
        
        self.model_args = kwargs
        
    def preprocess_data(self, X):
        
        if not hasattr(self, 'preprocessor'):
            self.preprocessor = make_column_transformer(
                (
                    OneHotEncoder(sparse_output=False, handle_unknown="ignore", drop="if_binary"),
                    X.select_dtypes(include="category").columns
                ),
                (
                    make_pipeline(
                        StandardScaler(),
                        SimpleImputer(strategy="constant", fill_value=0)
                    ),
                    X.select_dtypes(include="number").columns
                ),
                remainder="passthrough",
                verbose_feature_names_out=False
            ).set_output(transform="pandas")
            
            X = self.preprocessor.fit_transform(X)
            
        else:
            X = self.preprocessor.transform(X)
            
        self.n_feats = X.shape[1]
        return X    
        
    
    def get_data_loader(self, X, y, pairs=None, shuffle=True):
        
        assert "time" in y.dtype.names and "event" in y.dtype.names, \
            "y must be a structured array with 'time' and 'event' fields."
        
        # Convert X to numpy if pandas
        if hasattr(X, 'values'):
            X = X.values
        if pairs is not None:
            if hasattr(pairs, 'values'):
                pairs = pairs.values

        if pairs is not None:
            dataset = torch.utils.data.TensorDataset(
                torch.FloatTensor(X), 
                torch.FloatTensor(pairs),
                torch.BoolTensor(y["event"]),
                torch.FloatTensor(y["time"].copy()),
            )
        else:
            dataset = torch.utils.data.TensorDataset(
                torch.FloatTensor(X), 
                torch.BoolTensor(y["event"]),
                torch.FloatTensor(y["time"].copy()),
            )
        loader = torch.utils.data.DataLoader(dataset, batch_size=self.batch_size, shuffle=shuffle)
        
        return loader
    
    def train_epoch(self, model, train_loader, optimizer):
        model.train()
        total_loss = 0

        for X_main, events, times in tqdm(train_loader, leave=False):

            X_main, events, times = X_main.to(self.device), events.to(self.device), times.to(self.device)

            y_pred = model(X_main)
            y_pred = torch.sigmoid(y_pred).squeeze(-1)

            loss = drsa_loss(y_pred, events, times, self.eval_times)
            
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()

            total_loss += loss.item()
            
        return total_loss / len(train_loader)
    
    def test_epoch(self, model, test_loader):
        model.eval()
        total_loss = 0
        preds = []
        
        with torch.no_grad():
            for X_main, events, times in tqdm(test_loader, leave=False):
                X_main, events, times = X_main.to(self.device), events.to(self.device), times.to(self.device)

                y_pred = model(X_main)
        
                preds.append(y_pred.detach().squeeze(-1))
            
                y_pred = model(X_main)
                y_pred = torch.sigmoid(y_pred).squeeze(-1)
                
                loss = drsa_loss(y_pred, events, times, self.eval_times)
                
                # surv_probs = torch.cumprod(1 - y_pred, dim=1)

                # evt_loss = event_time_loss(y_pred)
                # evr_loss = event_rate_loss(y_pred)
                # loss = (alpha * evt_loss) + ((1 - alpha) * evr_loss)
                
                # loss = bce_surv_loss(surv_probs, events, times, eval_times)

                total_loss += loss.item() 
        
        return total_loss / len(test_loader), torch.cat(preds)
    
    def fit_one_split(self, X_train, y_train, X_val, y_val):
        # If selected_feats is set, only use those features
        if hasattr(self, 'selected_feats'):
            X_train = X_train[self.selected_feats]
            X_val = X_val[self.selected_feats]
        
        # Get data loaders
        train_loader = self.get_data_loader(X_train, y_train)
        val_loader = self.get_data_loader(X_val, y_val, shuffle=False)
        
        model = DRSASingleSplit(
            n_input=self.n_input,
            n_hidden=self.n_hidden,
            n_output=self.n_eval_times,
            eval_times=self.eval_times,
            device=self.device
        ).to(self.device)
        
        
        optimizer = torch.optim.Adam(model.parameters(), lr=self.learning_rate)
        model.train_(
            train_epoch_fn=self.train_epoch,
            test_epoch_fn=self.test_epoch,
            train_loader=train_loader,
            test_loader=val_loader,
            optimizer=optimizer,
            n_epochs=self.max_epochs
        )
        
        return model
    
    def fit(self, X, y):
        
        # Preprocess features
        X = self.preprocess_data(X)
        
        self.feature_names_in_ = X.columns
        self.n_input = X.shape[1]
        
        # Get evaluation times before train/val split
        quantiles = torch.quantile(
            torch.FloatTensor(y["time"].copy()),
            torch.linspace(0, 1, self.n_eval_times+2)
        )

        self.eval_times = quantiles[1:-1].to(self.device)
        
        # Remove duplicates in eval times
        if len(self.eval_times.unique()) < len(self.eval_times):
            self.eval_times = self.eval_times.unique()
            self.n_eval_times = len(self.eval_times)
        
        
        # Fit several models, one for each validation split
        self.models = []
        for i in range(self.n_val_splits):
            print("SPlIT", i)
        
            # Split the data into training and validation
            X_train, X_val, y_train, y_val = train_test_split(X, y, test_size=self.validation_size, random_state=10+i)
            
            # Fit to this split
            model = self.fit_one_split(X_train, y_train, X_val, y_val)
            model.feature_names_in_ = X_train.columns   
            
            self.models.append(model)
            
        return
    
    def predict(self, X_test):
        
        # Preprocess features
        X_test = self.preprocess_data(X_test)
        
        # Make placeholder y_test
        y_test = np.zeros(X_test.shape[0], dtype=[("event", "?"), ("time", "f8")])
        
        test_loader = self.get_data_loader(X_test, y_test, shuffle=False)
        
        test_preds = np.zeros((self.n_val_splits, X_test.shape[0], len(self.eval_times)))
        for i, model in enumerate(self.models):
            _, model_preds = self.test_epoch(model, test_loader)
            test_preds[i, ...] = model_preds.cpu().numpy()
            
        return np.mean(test_preds, axis=0)
    
    def get_calibration_data(self, X, y, eval_time, n_bins=20, method="quantile"):
        
        # First get cdf preds
        preds = self.predict(X)
        hazard_preds = 1 / (1 + np.exp(-preds))
        cdf_preds = 1 - np.cumprod(1 - hazard_preds, axis=1) 
        eval_index = np.searchsorted(self.eval_times.cpu().numpy(), eval_time)
        cdf_preds = cdf_preds[:, eval_index]
        
        
        if method == "quantile":
            # Get bins from quantiles of predictions
            quantiles = np.quantile(
                cdf_preds,
                np.linspace(0, 1, n_bins+2)
            )
        elif method == "uniform":
            # Get bins from uniform spacing
            quantiles = np.linspace(cdf_preds.min(), cdf_preds.max(), n_bins+2)
        else:
            raise ValueError("method must be 'quantile' or 'uniform'")
        
        predicted = []
        observed = []
        
        for bin_ in zip(quantiles[:-1], quantiles[1:]):
            
            if bin_[0] == bin_[1]:
                bin_data = cdf_preds[
                    cdf_preds == bin_[0]
                ]
                bin_y = y[
                    cdf_preds == bin_[0]
                ]
                
            else:
                bin_data = cdf_preds[
                    np.logical_and(
                        cdf_preds >= bin_[0],
                        cdf_preds < bin_[1]
                    )
                ]
                bin_y = y[
                    np.logical_and(
                        cdf_preds >= bin_[0],
                        cdf_preds < bin_[1]
                    )
                ]

            
            times, surv_prob = kaplan_meier_estimator(bin_y["event"], bin_y["time"])
            predicted.append(bin_data.mean())
            observed.append(
                1 - surv_prob[np.clip(
                    np.searchsorted(times, eval_time), 0, len(times)-1
                )]
            )
            
            
        return np.array(predicted), np.array(observed), quantiles


import sys
sys.path.append("../")

import numpy as np
import pandas as pd
from sksurv.metrics import cumulative_dynamic_auc, concordance_index_censored, integrated_brier_score
from sklearn.model_selection import train_test_split
from tqdm import tqdm
import argparse
import torch
import torch.nn as nn
import torch.nn.functional as F
import pickle
import yaml
from utils import get_dataset
from sklearn.compose import make_column_transformer
from sklearn.preprocessing import OneHotEncoder, StandardScaler
from sklearn.impute import SimpleImputer
from sklearn.pipeline import make_pipeline
from sksurv.linear_model import CoxPHSurvivalAnalysis

sys.path.append('../models')
from coxnam import CoxNAM

device = "cuda:4" if torch.cuda.is_available() else "cpu"

print("DEVICE:", device)

parser = argparse.ArgumentParser()
parser.add_argument(
    '--dataset_name', 
    type=str, 
    help='Name of the dataset to process'
)
parser.add_argument(
    '--seed', 
    type=int, 
    default=10, 
    help='Seed to use for random number generation'
)
parser.add_argument(
    "--n_eval_times",
    type=int,
    default=100,
    help="Number of discrete times used during evaluation."
)
args = parser.parse_args()

print("GETTING DATSET...")
X, y = get_dataset(args.dataset_name)

# Do train/test split
X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=args.seed)

preprocessor = make_column_transformer(
    (
        OneHotEncoder(sparse_output=False, handle_unknown="ignore", drop="if_binary"),
        X.select_dtypes(include="category").columns
    ),
    (
        make_pipeline(
            StandardScaler(),
            SimpleImputer(strategy="constant", fill_value=0)
        ),
        X.select_dtypes(include="number").columns
    ),
    remainder="passthrough",
    verbose_feature_names_out=False
).set_output(transform="pandas")

X_train = preprocessor.fit_transform(X_train)
X_test = preprocessor.transform(X_test)

model = CoxPHSurvivalAnalysis(alpha=1e-2)

from time import time  

fit_start_time = time()
model.fit(X_train, y_train)
fit_time = time() - fit_start_time
model.fit_time = fit_time

# Generate unique id to save with parameters
import random
param_id = random.randint(100000, 999999)


# Save run parameters to file
with open(f"../run_parameters/coxph_{args.dataset_name}_seed{args.seed}_params{param_id}.yaml", "w") as f:
    yaml.dump(vars(args), f)  
    
    
print("Saving model...")
model.params_id = param_id
with open(f"../model_saves/coxph_{args.dataset_name}_seed{args.seed}_params{param_id}.pkl", "wb") as f:
    pickle.dump(model, f)

print("Evaluating...")

quantiles = torch.quantile(
    torch.FloatTensor(y_train["time"].copy()),
    torch.linspace(0, 1, args.n_eval_times+2)
)

test_times = quantiles[1:-1].cpu().numpy()

# Remove duplicates in eval times
if len(np.unique(test_times)) < len(test_times):
    test_times = np.unique(test_times)
    
# Clip to y_test min and max
test_times = np.unique(np.clip(
    test_times, 
    max(y_train["time"].min(), y_test[y_test["event"]]["time"].min()) + 1e-4, 
    min(y_train["time"].max(), y_test[y_test["event"]]["time"].max()) - 1e-4
))

surv_fns = model.predict_survival_function(X_test)
cdf_preds = 1 - np.array([fn(test_times) for fn in surv_fns])

risk_preds = -1 * np.log(np.clip(1 - cdf_preds, 1e-5, 10 - 1e-5))

# Get time-dependent AUC
aucs, mean_auc = cumulative_dynamic_auc(y_train, y_test, risk_preds, test_times)

# Get normal c-index (Harrell)
cindex = concordance_index_censored(
    y_test["event"],    
    y_test["time"],
    model.predict(X_test)
)[0]

# Get brier score
brier_score = integrated_brier_score(
    y_train,
    y_test,
    1 - cdf_preds,
    test_times
)

results = pd.DataFrame(
    [["CoxPH", args.dataset_name, args.seed, cindex, mean_auc, brier_score, param_id]],
    columns=["model", "dataset", "seed", "cindex", "mean_auc", "brier_score", "param_id"]
)

results.to_csv(
    f"../metric_saves/coxph_{args.dataset_name}_seed{args.seed}.csv", index=False
)


import logging

def get_logger(name: str, verbosity: int = 0):
    """
    Create a logger for a specific class or function with the given verbosity level.
    
    Parameters:
    name (str): Name of the logger.
    verbosity (int): Verbosity level (0 = None, 1 = Info, 2 = Debug).

    Returns:
    logging.Logger: Configured logger.
    """
    logger = logging.getLogger(name)
    
    if verbosity == 0:
        logger.setLevel(logging.WARNING)  # Only warnings and errors
    elif verbosity == 1:
        logger.setLevel(logging.INFO)     # General progress info
    elif verbosity >= 2:
        logger.setLevel(logging.DEBUG)    # Debugging info
    
    # Clear existing handlers to avoid duplication
    if not logger.handlers:
        handler = logging.StreamHandler()
        formatter = logging.Formatter('[%(levelname)s] %(message)s')
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        logger.propagate = False

    return logger

class LoggingMixin:
    def __init__(self, verbosity=0):
        self.logger = get_logger(self.__class__.__name__, verbosity)




import numpy as np 
import pandas as pd 
from itertools import combinations
from sklearn.model_selection import train_test_split
import torch
from dnamite.models import DNAMiteSurvival
from sksurv.datasets import load_flchain

def get_data():
    X, y = load_flchain()

    # Change y dtypes and names
    y.dtype = np.dtype([('event', bool), ('time', float)])

    # # Impute chapter with new category
    X["chapter"] = X["chapter"].astype("category")
    X["chapter"] = X["chapter"].cat.add_categories("Unknown")
    X["chapter"] = X["chapter"].fillna("Unknown")

    # Clip Training data to avoid overflow errors
    X["creatinine"] = X["creatinine"].clip(-5, 5)

    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=10)
    return X_train, X_test, y_train, y_test

def test_init():
    X_train, X_test, y_train, y_test = get_data()
    
    # Minimal initialization
    model = DNAMiteSurvival(
        n_features=X_train.shape[1],
    )
    assert model.kernel_size > 0
    assert model.kernel_weight > 0
    assert model.pair_kernel_weight > 0
    assert model.pair_kernel_size > 0
    assert model.verbosity == 0
    assert model.n_output == 100
    
def test_train():
    X_train, X_test, y_train, y_test = get_data()
    
    # First fit without pairs
    model = DNAMiteSurvival(
        n_features=X_train.shape[1],
        max_epochs=2,
        n_val_splits=2,
    )
    model.fit(X_train, y_train)
    preds = model.predict(X_test)
    assert preds is not None
    
    # Then fit with pairs
    model = DNAMiteSurvival(
        n_features=X_train.shape[1],
        max_epochs=2,
        n_val_splits=2,
        num_pairs=3
    )
    model.fit(X_train, y_train)
    for m in model.models:
        assert m.n_pairs == 3
        assert m.pairs_list is not None
        assert m.selected_pair_indices is not None
        
    preds = model.predict(X_test)
    assert preds is not None


def test_train_with_pairs_list():
    
    X_train, X_test, y_train, y_test = get_data()
    model = DNAMiteSurvival(
        n_features=X_train.shape[1],
        pairs_list=[(0, 1), (1, 2), (2, 3)],
        max_epochs=2,
        n_val_splits=2
    )
    assert model.pairs_list == [(0, 1), (1, 2), (2, 3)]
    
    model.fit(X_train, y_train)
    assert model.pairs_list == [(0, 1), (1, 2), (2, 3)]
    for m in model.models:
        assert m.selected_pair_indices == [(0, 1), (1, 2), (2, 3)]
        assert torch.equal(m.pairs_list, torch.tensor([(0, 1), (1, 2), (2, 3)]))
        assert m.n_pairs == 3
        
    preds = model.predict(X_test)
    assert preds is not None
    
def test_select_features():
    X_train, X_test, y_train, y_test = get_data()
    model = DNAMiteSurvival(
        n_features=X_train.shape[1],
        n_val_splits=2,
        max_epochs=6,
    )
    model.select_features(X_train, y_train, reg_param=0.2, gamma=0.04)
    assert len(model.selected_feats) < model.n_features
    assert model.pairs_list is None
    model.fit(X_train, y_train)
    assert len(model.selected_feats) < model.n_features
    assert model.pairs_list is None
    preds = model.predict(X_test)
    assert preds is not None
    
    # Now try with also selecting pairs
    model = DNAMiteSurvival(
        n_features=X_train.shape[1],
        n_val_splits=2,
        max_epochs=6,
    )
    model.select_features(X_train, y_train, select_pairs=True, reg_param=0.2, gamma=0.04, pair_reg_param=0.005, pair_gamma=0.02)
    assert len(model.selected_feats) < model.n_features
    assert model.pairs_list is not None
    assert len(model.pairs_list) < model.n_features * (model.n_features - 1) // 2
    assert model.selected_pairs == [[model.feature_names_in_[pair[0]], model.feature_names_in_[pair[1]]] for pair in model.pairs_list]
    model.fit(X_train, y_train)
    assert len(model.selected_feats) < model.n_features
    assert model.pairs_list is not None
    assert len(model.pairs_list) < model.n_features * (model.n_features - 1) // 2
    assert model.selected_pairs == [[model.feature_names_in_[pair[0]], model.feature_names_in_[pair[1]]] for pair in model.pairs_list]
    preds = model.predict(X_test)
    assert preds is not None
    
def test_feature_importances():
    X_train, _, y_train, _ = get_data()
    model = DNAMiteSurvival(
        n_features=X_train.shape[1],
        n_val_splits=2,
        max_epochs=2,
    )
    model.fit(X_train, y_train)
    feat_imps = model.get_feature_importances(eval_time=365)
    assert feat_imps["importance"].min() > 0
    model.plot_feature_importances(n_features=5, eval_times=365, missing_bin="include")
    model.plot_feature_importances(n_features=5, eval_times=365, missing_bin="ignore")
    model.plot_feature_importances(n_features=5, eval_times=365, missing_bin="stratify")
    
    # Now try with also selecting pairs
    model = DNAMiteSurvival(
        n_features=X_train.shape[1],
        n_val_splits=2,
        max_epochs=2,
        num_pairs=3
    )
    model.fit(X_train, y_train)
    feat_imps = model.get_feature_importances(eval_time=365)
    assert feat_imps["importance"].min() > 0
    model.plot_feature_importances(n_features=5, eval_times=365, missing_bin="include")
    model.plot_feature_importances(n_features=5, eval_times=365, missing_bin="ignore")
    model.plot_feature_importances(n_features=5, eval_times=365, missing_bin="stratify")
    
def test_shape_functions():
    X_train, _, y_train, _ = get_data()
    model = DNAMiteSurvival(
        n_features=X_train.shape[1],
        n_val_splits=2,
        max_epochs=2,
    )
    model.fit(X_train, y_train)
    _ = model.get_shape_function("chapter", eval_time=365)
    _ = model.get_shape_function("creatinine", eval_time=365)
    model.plot_shape_function("chapter", eval_times=365)
    model.plot_shape_function(["chapter", "creatinine"], eval_times=365)
    
    # Now try with also selecting pairs
    model = DNAMiteSurvival(
        n_features=X_train.shape[1],
        n_val_splits=2,
        max_epochs=2,
        num_pairs=3
    )
    model.fit(X_train, y_train)
    
    example_pair = model.pairs_list[0]
    feat1, feat2 = model.feature_names_in_[example_pair[0]], model.feature_names_in_[example_pair[1]]
    _ = model.get_pair_shape_function(feat1, feat2, eval_time=365)
    model.plot_pair_shape_function(feat1, feat2, eval_times=365)
    
def test_survival_preds():
    X_train, X_test, y_train, y_test = get_data()
    model = DNAMiteSurvival(
        n_features=X_train.shape[1],
        n_val_splits=2,
        max_epochs=2,
    )
    model.fit(X_train, y_train)
    surv_preds = model.predict_survival(X_test)
    assert surv_preds.shape[0] == X_test.shape[0]
    assert surv_preds.shape[1] == len(model.eval_times)
    assert np.all((0 <= surv_preds) & (surv_preds <= 1))
    
    test_times = np.array([365, 730, 1095])
    surv_preds = model.predict_survival(X_test, test_times)
    assert surv_preds.shape[0] == X_test.shape[0]
    assert surv_preds.shape[1] == len(test_times)
    assert np.all((0 <= surv_preds) & (surv_preds <= 1))
    
def test_calibration():
    X_train, X_test, y_train, y_test = get_data()
    model = DNAMiteSurvival(
        n_features=X_train.shape[1],
        n_val_splits=2,
        max_epochs=2,
    )
    model.fit(X_train, y_train)
    predicted, observed, quantiles = model.get_calibration_data(X_test, y_test, eval_time=365, n_bins=10)
    assert len(predicted) == len(observed) == 10
    
    model.make_calibration_plot(X_test, y_test, eval_times=365, n_bins=10)
    
def test_kernel():
    X_train, X_test, y_train, y_test = get_data()
    
    # Test with no kernel size or weight
    model = DNAMiteSurvival(
        n_features=X_train.shape[1],
        n_val_splits=2,
        max_epochs=2,
        kernel_size=0,
        kernel_weight=0
    )
    model.fit(X_train, y_train)
    _ = model.predict(X_test)

def test_partial():
    X_train, X_test, y_train, y_test = get_data()
    
    # Test with normal
    model = DNAMiteSurvival(
        n_features=X_train.shape[1],
        n_val_splits=2,
        max_epochs=2,
    )
    model.fit(X_train, y_train, partialed_feats=[X_train.columns[0]])
    _ = model.predict(X_test)
    
    # Test with pairs
    model = DNAMiteSurvival(
        n_features=X_train.shape[1],
        n_val_splits=2,
        max_epochs=2,
        num_pairs=3
    )
    model.fit(X_train, y_train, partialed_feats=[X_train.columns[0]])
    _ = model.predict(X_test)
    
    # Test with feature selection
    model = DNAMiteSurvival(
        n_features=X_train.shape[1],
        n_val_splits=2,
        max_epochs=10,
        batch_size=32
    )
    model.select_features(X_train, y_train, reg_param=0.2, gamma=0.04, partialed_feats=[X_train.columns[0]])
    model.fit(X_train, y_train, partialed_feats=[X_train.columns[0]])
    _ = model.predict(X_test)
    
def test_monotone():
    X_train, X_test, y_train, y_test = get_data()
    
    # Test with normal
    model = DNAMiteSurvival(
        n_features=X_train.shape[1],
        n_val_splits=2,
        max_epochs=2,
        monotone_constraints=[1] + [0] * (X_train.shape[1] - 1)
    )
    model.fit(X_train, y_train)
    
    shape_data = model.get_shape_function(X_train.columns[0], eval_time=365)
    for split in shape_data["split"].unique():
        split_shape_data = shape_data[(shape_data["split"] == split) & (shape_data["bin"].notna())]
        split_shape_data = split_shape_data.sort_values("bin")
        assert split_shape_data["score"].diff().dropna().ge(0).all()
    
    


"""
SAMPLE BEGINNING OF FILE DOCSTRING
"""

import torch
import torch.nn as nn
import numpy as np
import pandas as pd
from tqdm import tqdm
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import OneHotEncoder, StandardScaler
from sklearn.impute import SimpleImputer
from sklearn.compose import make_column_transformer
from sklearn.pipeline import make_pipeline
from .nam import BaseSingleSplitNAM
from sksurv.nonparametric import kaplan_meier_estimator
from sksurv.linear_model.coxph import BreslowEstimator
from dnamite.loss_fns import coxph_loss


class CoxNAM(nn.Module):
    """ 
    CoxNAM
    """
    
    def __init__(
        self, 
        n_features, 
        n_hidden, 
        validation_size=0.2,
        n_val_splits=5,
        learning_rate=1e-4,
        max_epochs=100,
        batch_size=128,
        device="cpu",
        **kwargs
    ):
        super().__init__()
        self.n_features = n_features
        self.n_hidden = n_hidden
        self.validation_size = validation_size
        self.n_val_splits = n_val_splits
        self.learning_rate = learning_rate
        self.max_epochs = max_epochs
        self.batch_size = batch_size
        self.device = device
        
        self.model_args = kwargs
        
    def preprocess_data(self, X):
        
        if not hasattr(self, 'preprocessor'):
            self.preprocessor = make_column_transformer(
                (
                    OneHotEncoder(sparse_output=False, handle_unknown="ignore", drop="if_binary"),
                    X.select_dtypes(include="category").columns
                ),
                (
                    make_pipeline(
                        StandardScaler(),
                        SimpleImputer(strategy="constant", fill_value=0)
                    ),
                    X.select_dtypes(include="number").columns
                ),
                remainder="passthrough",
                verbose_feature_names_out=False
            ).set_output(transform="pandas")
            
            X = self.preprocessor.fit_transform(X)
            
        else:
            X = self.preprocessor.transform(X)
            
        return X    
        
    
    def get_data_loader(self, X, y, pairs=None, shuffle=True):
        
        assert "time" in y.dtype.names and "event" in y.dtype.names, \
            "y must be a structured array with 'time' and 'event' fields."
        
        # Convert X to numpy if pandas
        if hasattr(X, 'values'):
            X = X.values
            
        dataset = torch.utils.data.TensorDataset(
            torch.FloatTensor(X), 
            torch.BoolTensor(y["event"]),
            torch.FloatTensor(y["time"].copy()),
        )
        loader = torch.utils.data.DataLoader(dataset, batch_size=self.batch_size, shuffle=shuffle)
        
        return loader
    
    def train_epoch_mains(self, model, train_loader, optimizer):
        model.train()
        total_loss = 0

        for X_main, events, times in tqdm(train_loader, leave=False):

            X_main, events, times = X_main.to(self.device), events.to(self.device), times.to(self.device)

            y_pred = model(mains=X_main).squeeze()
            
            loss = coxph_loss(
                y_pred,
                events,
                times
            )
            
            if model.penalized:
                loss += model.loss_penalty()

            optimizer.zero_grad()
            loss.backward()
            optimizer.step()

            total_loss += loss.item()
            
        return total_loss / len(train_loader)
    
    def test_epoch_mains(self, model, test_loader):
        model.eval()
        total_loss = 0
        preds = []
        
        with torch.no_grad():
            for X_main, events, times in tqdm(test_loader, leave=False):

                X_main, events, times = X_main.to(self.device), events.to(self.device), times.to(self.device)

                y_pred = model(mains=X_main).squeeze()
        
                preds.append(y_pred.detach())
            
                loss = coxph_loss(
                    y_pred,
                    events,
                    times
                )
                
                if model.penalized:
                    loss += model.loss_penalty()
                    
                total_loss += loss.item() 
        
        return total_loss / len(test_loader), torch.cat(preds)
    
    def fit_one_split(self, X_train, pseudo_vals_train, X_val, pseudo_vals_val):
        # If selected_feats is set, only use those features
        if hasattr(self, 'selected_feats'):
            X_train = X_train[self.selected_feats]
            X_val = X_val[self.selected_feats]
        
        # Get data loaders
        train_loader = self.get_data_loader(X_train, pseudo_vals_train)
        val_loader = self.get_data_loader(X_val, pseudo_vals_val, shuffle=False)
        
        model = BaseSingleSplitNAM(
            n_features=self.n_features, 
            n_hidden=self.n_hidden, 
            n_output=1,
            device=self.device,
            **self.model_args
        ).to(self.device)
        
        optimizer = torch.optim.Adam(model.parameters(), lr=self.learning_rate)
        model.train_(
            train_epoch_fn=self.train_epoch_mains,
            test_epoch_fn=self.test_epoch_mains,
            train_loader=train_loader,
            test_loader=val_loader,
            optimizer=optimizer,
            n_epochs=self.max_epochs
        )
        
        return model
    
    def fit(self, X, y, fit_baseline=True):
        
        # Preprocess features
        X_processed = self.preprocess_data(X)
        
        self.feature_names_in_ = X_processed.columns
        self.n_features = X_processed.shape[1]
        
        # Fit several models, one for each validation split
        self.models = []
        for i in range(self.n_val_splits):
            print("SPlIT", i)
        
            # Split the data into training and validation
            X_train, X_val, y_train, y_val = train_test_split(X_processed, y, test_size=self.validation_size, random_state=10+i)
            
            # Fit to this split
            model = self.fit_one_split(X_train, y_train, X_val, y_val)
            model.feature_names_in_ = X_train.columns   
            
            self.models.append(model)
            
        if fit_baseline:
            self.fit_baseline(X, y)
            
        return
    
    def predict(self, X_test):
        
        # Preprocess features
        X_test = self.preprocess_data(X_test)
        
        # Make placeholder y_test
        y_test = np.zeros(X_test.shape[0], dtype=[("event", "?"), ("time", "f8")])
        
        test_loader = self.get_data_loader(X_test, y_test, shuffle=False)
        
        test_preds = np.zeros((self.n_val_splits, X_test.shape[0]))
        for i, model in enumerate(self.models):
            _, model_preds = self.test_epoch_mains(model, test_loader)
            test_preds[i, ...] = model_preds.cpu().numpy()
            
        return np.mean(test_preds, axis=0)
    
    def fit_baseline(self, X, y):
        
        self.breslow = BreslowEstimator()
        pred = self.predict(X)
        self.breslow.fit(pred, y["event"], y["time"])
        
        return
    
    def predict_survival(self, X, eval_times):
        
        if not hasattr(self, 'breslow'):
            raise ValueError("Must fit baseline model first.")
        
        pred = self.predict(X)
        surv_fns = self.breslow.get_survival_function(pred)
        return np.array([fn(eval_times) for fn in surv_fns])

    
    def get_shape_function(self, feature_name, X, eval_time, feat_min=None, feat_max=None):
        
        # placeholder y
        y = np.zeros(X.shape[0], dtype=[("event", "?"), ("time", "f8")])
        
        X = self.preprocess_data(X)
        
        dfs = []
        for i, model in enumerate(self.models):
            
            print("SPLIT", i)
            print("X shape", X.shape)
            
            X_train, _, y_train, _ = train_test_split(X, y, test_size=self.validation_size, random_state=10+i)
            
            print("X_train shape", X_train.shape)
            
            train_loader = self.get_data_loader(X_train, y_train, shuffle=False)
            print("TRAIN LOADER", next(iter(train_loader))[0].shape)
            
            
            model.compute_intercept(train_loader)
            
            feat_index = model.feature_names_in_.get_loc(feature_name)
            
            if feat_min is None:
                input_min = X_train.iloc[:, feat_index].min()
                input_max = X_train.iloc[:, feat_index].max()
            else:
                scaler = self.preprocessor.transformers_[1][1][0]
                feat_index_in_scaler = scaler.feature_names_in_.tolist().index(feature_name)
                feat_mean = scaler.mean_[feat_index_in_scaler]
                feat_std = scaler.scale_[feat_index_in_scaler]
                input_min = (feat_min - feat_mean) / feat_std
                input_max = (feat_max - feat_mean) / feat_std
            
            feat_shape, feat_inputs = model.get_shape_function(feat_index, input_min, input_max, center=True)
            
            if feat_min is not None:
                feat_inputs = feat_inputs * feat_std + feat_mean
            
            dfs.append(
                pd.DataFrame({
                    "feature": feature_name,
                    "shape": feat_shape.squeeze().cpu().numpy(),
                    "input": feat_inputs.cpu().numpy().round(3),
                    "split": i
                })
            )
            
        return pd.concat(dfs)
    
    def get_calibration_data(self, X, y, eval_time, n_bins=20, method="quantile"):
        
        # First get cdf preds
        surv_preds = self.predict_survival(X, eval_time)
        cdf_preds = 1 - surv_preds
        
        
        if method == "quantile":
            # Get bins from quantiles of predictions
            quantiles = np.quantile(
                cdf_preds,
                np.linspace(0, 1, n_bins+2)
            )
        elif method == "uniform":
            # Get bins from uniform spacing
            quantiles = np.linspace(cdf_preds.min(), cdf_preds.max(), n_bins+2)
        else:
            raise ValueError("method must be 'quantile' or 'uniform'")
        
        predicted = []
        observed = []
        
        for bin_ in zip(quantiles[:-1], quantiles[1:]):
            
            if bin_[0] == bin_[1]:
                bin_data = cdf_preds[
                    cdf_preds == bin_[0]
                ]
                bin_y = y[
                    cdf_preds == bin_[0]
                ]
                
            else:
                bin_data = cdf_preds[
                    np.logical_and(
                        cdf_preds >= bin_[0],
                        cdf_preds < bin_[1]
                    )
                ]
                bin_y = y[
                    np.logical_and(
                        cdf_preds >= bin_[0],
                        cdf_preds < bin_[1]
                    )
                ]

            
            times, surv_prob = kaplan_meier_estimator(bin_y["event"], bin_y["time"])
            predicted.append(bin_data.mean())
            observed.append(
                1 - surv_prob[np.clip(
                    np.searchsorted(times, eval_time), 0, len(times)-1
                )]
            )
            
            
        return np.array(predicted), np.array(observed), quantiles


import torch


def ipcw_rps_loss(
    cdf_preds, 
    pcw_eval_times, 
    pcw_obs_times,
    events, 
    times, 
    eval_times
):
    """  
    IPCW Loss
    """
    
    # cdf_preds is of shape (N, T) where N is batch size and T is len(eval_times)
    #     cdf_preds[i, j] is hat(P)(Ti <= tj | Xi)
    # pcw_eval_times is of shape (T)
    #     pcw_eval_times[j] is hat(P)(C > tj)
    # pcw_obs_times is of shape (N)
    #     pcw_obs_times[i] is hat(P)(C > Ti)
    
    # Maximize survival before event time
    loss = torch.sum(
        (cdf_preds)**2 * \
        (eval_times < times.unsqueeze(-1)).float() / \
        pcw_eval_times.repeat(cdf_preds.shape[0], 1),
        axis=1
    )
    
    # Minimize survival past event time for patients that are not censored
    loss += torch.sum(
        (1 - cdf_preds)**2 * \
        torch.logical_and(
            eval_times >= times.unsqueeze(-1),
            events.unsqueeze(-1)
        ).float() / \
        pcw_obs_times.unsqueeze(-1).repeat(1, cdf_preds.shape[1]),
        axis=1
    )
    
    # Normalize by censor distribution to get correct convergence
    # divisor = (eval_times < censor_times.unsqueeze(-1)).sum(axis=0) + 1e-5
    
    # return torch.sum(loss / divisor)
    return torch.mean(loss)

def rps_loss(surv_preds, events, times, eval_times):
    """  
    RPS Loss
    """
    
    # surv_preds is of shape (N, T)
    # where N is batch size and T is number of evaluation times
    # surv_preds(i, j) gives P(Ti > tj | Xi)
    
    uncensored_loss = torch.square(surv_preds - (eval_times < times.unsqueeze(-1)).float()).sum(axis=1)
    censored_loss = (torch.square(surv_preds - 1) * (eval_times <= times.unsqueeze(-1)).float()).sum(axis=1)
    
    return torch.mean(events.float() * uncensored_loss + (1 - events.float()) * censored_loss)

# Loss taken from Eq 1 of https://dl.acm.org/doi/pdf/10.1145/3534678.3539259
# Can't just use MSE loss because pseudo-values not strictly in [0, 1]
def pseudo_value_loss(
    surv_preds,
    pseudo_values
):
    """  
    pseudo-value loss
    """
    
    return torch.mean(
        pseudo_values * (1 - 2*surv_preds) + torch.square(surv_preds), dim=0
    ).sum()
    
def coxph_loss(y_hat, events, times):
    """
    CoxPH Loss
    """
    
    # Sort y_hat and events by times
    sort_idx = torch.argsort(times)
    y_hat = y_hat[sort_idx]
    events = events[sort_idx]
    
    e = torch.exp(y_hat)
    e_sums = torch.cumsum(e.flip(0), 0).flip(0)
    
    
    return -1 * events.to(torch.float32) @ (y_hat - torch.log(e_sums))

def bce_surv_loss(surv_preds, events, times, eval_times):
    """
    Binary Cross-Entropy Survival Loss
    """
    
    # surv_preds is of shape (N, T)
    # where N is batch size and T is number of evaluation times
    # surv_preds(i, j) gives P(Ti > tj | Xi)
    
    # Maximize survival for all samples before their times
    loss = torch.sum(
        -1 * torch.log(surv_preds + 1e-5) * \
        (eval_times < times.unsqueeze(-1)).float(),
        axis=1
    )
    
    # Minimize survival loss for all uncensored samples after their times
    loss += torch.sum(
        -1 * torch.log(1 - surv_preds + 1e-5) * \
        torch.logical_and(
            eval_times >= times.unsqueeze(-1),
            events.unsqueeze(-1)
        ).float(),
        axis=1
    )
    
    return torch.mean(loss)

def drsa_loss(hazard_preds, events, times, eval_times):
    """
    DRSA-Loss
    """
    
    # Find evaluation time for each event time
    sample_eval_times = torch.clip(
        torch.searchsorted(eval_times, times).unsqueeze(-1),
        0,
        len(eval_times)-1
    )
    
    
    # Get first part of loss
    loss = -1 * events.unsqueeze(-1).float() * (
        torch.log(hazard_preds + 1e-5).gather(1, sample_eval_times) + \
        torch.sum(torch.log(1 - hazard_preds + 1e-5) * (eval_times < times.unsqueeze(-1)).float(), dim=1)
    )
    
    # Get second part of loss
    
    surv_preds = torch.cumprod(1 - hazard_preds, dim=1)
    
    loss += -1 * (~events).float() * torch.log(surv_preds + 1e-5).gather(1, sample_eval_times)
    loss += -1 * events.float() * torch.log(1 - surv_preds + 1e-5).gather(1, sample_eval_times)
    
    return loss.mean()


import numpy as np
import pandas as pd
import subprocess

datasets = ["flchain", "metabric", "support", "unos", "heart_failure"]
batch_sizes = [128, 64, 128, 512, 512]
seeds = [10, 11, 12, 13, 14]
# seeds = [10]
models = ["PseudoNAM"]

for dataset, batch_size in zip(datasets, batch_sizes):
    for seed in seeds:
        for model in models:
            if model == "PseudoNAM" and dataset in ["heart_failure", "unos"]:
                continue
            print(f"Running {model} on {dataset} with seed {seed}")
            subprocess.run([
                "python", 
                f"run_{model.lower()}.py", 
                "--dataset_name", dataset, 
                "--seed", str(seed),
                "--lr", str(5e-4),
                "--batch_size", str(batch_size),
            ])


"""
SAMPLE BEGINNING OF FILE DOCSTRING
"""

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
import math
import copy
from torch.autograd import Variable
from torch.utils.data import Dataset
from operator import itemgetter

from tqdm import tqdm
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import OneHotEncoder, StandardScaler
from sklearn.impute import SimpleImputer
from sklearn.compose import make_column_transformer
from sklearn.pipeline import make_pipeline
from sksurv.nonparametric import kaplan_meier_estimator
from dnamite.loss_fns import bce_surv_loss

class TranDataset(Dataset):
    def __init__(self, features, labels, max_time, is_train=True):
        self.is_train = is_train
        self.data = []

        temp = []
        for feature, label in zip(features, labels):
            feature = torch.from_numpy(feature).float()
            duration, is_observed = label[0], label[1]
            temp.append([duration, is_observed, feature])
        sorted_temp = sorted(temp, key=itemgetter(0))

        if self.is_train:
            new_temp = sorted_temp
        else:
            new_temp = temp

        for duration, is_observed, feature in new_temp:
            if is_observed:
                mask = max_time * [1.]
                label = duration * [1.] + (max_time - duration) * [0.]
                feature = torch.stack(max_time * [feature])
                self.data.append([feature.cuda(), torch.tensor(duration).float().cuda(), torch.tensor(mask).float().cuda(), torch.tensor(label).cuda(), torch.tensor(is_observed).byte().cuda()])
            else:
                # NOTE plus 1 to include day 0
                mask = (duration + 1) * [1.] + (max_time - (duration + 1)) * [0.]
                label = max_time * [1.]
                feature = torch.stack(max_time * [feature])
                self.data.append([feature.cuda(), torch.tensor(duration).float().cuda(), torch.tensor(mask).float().cuda(), torch.tensor(label).cuda(), torch.tensor(is_observed).byte().cuda()])

    def __getitem__(self, index_a):
        if self.is_train:
            if index_a == len(self.data) - 1:
                index_b = np.random.randint(len(self.data))
            else:
                # NOTE self.data is sorted
                index_b = np.random.randint(index_a+1, len(self.data))
                
            batch = [ [self.data[index_a][i], self.data[index_b][i]] for i in range(len(self.data[index_a])) ]
        else:
            batch = self.data[index_a]
            
        print("BATCH SIZE", batch[0].shape)
        return batch

    def __len__(self):
        return len(self.data)


def clones(module, N):
    return nn.ModuleList([copy.deepcopy(module) for _ in range(N)])


def attention(query, key, value, mask=None, dropout=None):
    d_k = query.size(-1)
    scores = torch.matmul(query, key.transpose(-2, -1)) / math.sqrt(d_k)
    if mask is not None:
        scores = scores.masked_fill(mask == 0, -1e9)
    p_attn = F.softmax(scores, dim = -1)
    if dropout is not None:
        p_attn = dropout(p_attn)
    return torch.matmul(p_attn, value), p_attn


class MultiHeadedAttention(nn.Module):
    def __init__(self, h, d_model, dropout=0.1):
        super().__init__()
        assert d_model % h == 0
        self.d_k = d_model // h
        self.h = h
        self.linears = clones(nn.Linear(d_model, d_model), 4)
        self.attn = None
        self.dropout = nn.Dropout(p=dropout)
        
    def forward(self, query, key, value, mask=None):
        if mask is not None:
            mask = mask.unsqueeze(1).unsqueeze(1)
        nbatches = query.size(0)
        
        query, key, value = \
            [l(x).view(nbatches, -1, self.h, self.d_k).transpose(1, 2)
             for l, x in zip(self.linears, (query, key, value))]
        
        x, self.attn = attention(query, key, value, mask=mask, dropout=self.dropout)
        
        x = x.transpose(1, 2).contiguous().view(nbatches, -1, self.h * self.d_k)
        return self.linears[-1](x)


class PositionalEncoding(nn.Module):
    def __init__(self, d_model, dropout, max_len=5000):
        super().__init__()
        self.dropout = nn.Dropout(p=dropout)
        
        pe = torch.zeros(max_len, d_model)
        position = torch.arange(0, max_len).unsqueeze(1).float()
        div_term = torch.exp(torch.arange(0, d_model, 2).float() * -torch.tensor(math.log(10000.0) / d_model))
        pe[:, 0::2] = torch.sin(position * div_term)
        pe[:, 1::2] = torch.cos(position * div_term)
        pe = pe.unsqueeze(0)
        self.register_buffer('pe', pe)
        
    def forward(self, x):
        x = x + Variable(self.pe[:, :x.size(1)], requires_grad=False)
        return self.dropout(x)


class PositionwiseFeedForward(nn.Module):
    def __init__(self, d_model, d_ff, dropout=0.1):
        super().__init__()
        self.w_1 = nn.Linear(d_model, d_ff)
        self.w_2 = nn.Linear(d_ff, d_model)
        self.dropout = nn.Dropout(dropout)

    def forward(self, x):
        return self.w_2(self.dropout(F.relu(self.w_1(x))))


class LayerNorm(nn.Module):
    def __init__(self, features, eps=1e-6):
        super().__init__()
        self.a_2 = nn.Parameter(torch.ones(features))
        self.b_2 = nn.Parameter(torch.zeros(features))
        self.eps = eps

    def forward(self, x):
        mean = x.mean(-1, keepdim=True)
        std = x.std(-1, keepdim=True)
        return self.a_2 * (x - mean) / (std + self.eps) + self.b_2


class SublayerConnection(nn.Module):
    def __init__(self, size, dropout):
        super().__init__()
        self.norm = LayerNorm(size)
        self.dropout = nn.Dropout(dropout)

    def forward(self, x, sublayer):
        return x + self.dropout(sublayer(self.norm(x)))


# Initial embedding for raw input
class SrcEmbed(nn.Module):
    def __init__(self, input_dim, d_model):
        super().__init__()
        self.w = nn.Linear(input_dim, d_model)
        self.norm = LayerNorm(d_model)

    def forward(self, x):
        return self.norm(self.w(x))


# Final layer for the Transformer
class TranFinalLayer(nn.Module):
    def __init__(self, d_model):
        super().__init__()
        self.w_1 = nn.Linear(d_model, d_model // 2)
        self.norm = LayerNorm(d_model // 2)
        self.w_2 = nn.Linear(d_model // 2, 1)

    def forward(self, x):
        x = F.relu(self.w_1(x))
        x = self.norm(x)
        x = self.w_2(x)
        return torch.sigmoid(x.squeeze(-1))


class Encoder(nn.Module):
    def __init__(self, layer, N, d_model, dropout, num_features):
        super().__init__()
        self.src_embed = SrcEmbed(num_features, d_model)
        self.position_encode = PositionalEncoding(d_model, dropout)
        self.layers = clones(layer, N)
        self.norm = LayerNorm(layer.size)
        self.final_layer = TranFinalLayer(d_model)
        
    def forward(self, x, mask=None):
        x = self.position_encode(self.src_embed(x))
        for layer in self.layers:
            x = layer(x, mask)
        return self.final_layer(x)


class EncoderLayer(nn.Module):
    def __init__(self, size, self_attn, feed_forward, dropout):
        super().__init__()
        self.self_attn = self_attn
        self.feed_forward = feed_forward
        self.sublayer = clones(SublayerConnection(size, dropout), 2)
        self.size = size

    def forward(self, x, mask):
        x = self.sublayer[0](x, lambda x: self.self_attn(x, x, x, mask))
        return self.sublayer[1](x, self.feed_forward)

# Transformer model for survival analysis
# Based on this paper: https://proceedings.mlr.press/v146/hu21a/hu21a.pdf
class SATransformerSingleSplit(nn.Module):
    def __init__(self, *, n_feats, d_model, n_eval_times, n_layers, num_heads, d_ff, drop_prob, device="cpu"):
        super().__init__()
        
        self.n_eval_times = n_eval_times
        
        self.embed = SrcEmbed(n_feats, d_model)
        self.transformer = torch.nn.TransformerEncoder(
            torch.nn.TransformerEncoderLayer(
                d_model=d_model,
                nhead=num_heads,
                dim_feedforward=d_ff,
                dropout=drop_prob,
                batch_first=True
            ),
            n_layers
        )
        self.final_layer = TranFinalLayer(d_model)
        
        # Create the positional encoding
        self.pe = torch.zeros(n_eval_times, d_model)
        position = torch.arange(0, n_eval_times).unsqueeze(1).float()
        div_term = torch.exp(torch.arange(0, d_model, 2).float() * -torch.tensor(np.log(10000.0) / d_model))
        self.pe[:, 0::2] = torch.sin(position * div_term)
        self.pe[:, 1::2] = torch.cos(position * div_term)
        self.pe = self.pe.unsqueeze(0).to(device)
        
    def forward(self, x):
        
        # Embed features
        embeds = self.embed(x)
        
        # Stack embeddings and add positional encoding
        embeds = embeds.unsqueeze(1).repeat(1, self.n_eval_times, 1)
        embeds += self.pe
        
        # Pass through transformer
        output = self.transformer(embeds)
        
        # Return final layer output
        return self.final_layer(output)
    
    def train_(
        self, 
        train_epoch_fn, 
        test_epoch_fn,
        train_loader,
        test_loader,
        optimizer,
        n_epochs,
    ):
        
        early_stopping_counter = 0
        best_test_loss = float('inf')

        for epoch in range(n_epochs):
            train_epoch_fn(self, train_loader, optimizer)
            test_loss, test_preds = test_epoch_fn(self, test_loader)
            
            # print(f"Epoch {epoch+1} | Train loss: {train_loss:.3f} | Test loss: {test_loss:.3f} | Num Feats: {len([z for z in self.get_smooth_z() if z > 0])}")

            # Check if the test loss has improved
            if test_loss < best_test_loss:
                best_test_loss = test_loss
                early_stopping_counter = 0
                
                # Save the model at the best test loss
                torch.save(self.state_dict(), "../model_saves/tmp_best_model.pt")
                
            else:
                early_stopping_counter += 1

            # If test loss has not improved for 5 consecutive epochs, terminate training
            if early_stopping_counter >= 5:
                print(f"Early stopping at {epoch+1} epochs: Test loss has not improved for 5 consecutive epochs.")
                break
            
        # Load the model from the best test loss
        self.load_state_dict(torch.load("../model_saves/tmp_best_model.pt"))

        return
    
class SATransformer(nn.Module):
    """   
    SATransformer
    """
    
    def __init__(
        self, 
        d_model, 
        n_eval_times, 
        n_layers, 
        num_heads, 
        d_ff, 
        drop_prob,
        validation_size=0.2,
        n_val_splits=5,
        learning_rate=1e-4,
        max_epochs=100,
        batch_size=128,
        device="cpu",
        **kwargs
    ):
        super().__init__()
        self.d_model = d_model
        self.n_eval_times = n_eval_times
        self.n_layers = n_layers
        self.num_heads = num_heads
        self.d_ff = d_ff
        self.drop_prob = drop_prob
        self.validation_size = validation_size
        self.n_val_splits = n_val_splits
        self.learning_rate = learning_rate
        self.max_epochs = max_epochs
        self.batch_size = batch_size
        self.device = device
        
        self.model_args = kwargs
        
    def preprocess_data(self, X):
        
        if not hasattr(self, 'preprocessor'):
            self.preprocessor = make_column_transformer(
                (
                    OneHotEncoder(sparse_output=False, handle_unknown="ignore", drop="if_binary"),
                    X.select_dtypes(include="category").columns
                ),
                (
                    make_pipeline(
                        StandardScaler(),
                        SimpleImputer(strategy="constant", fill_value=0)
                    ),
                    X.select_dtypes(include="number").columns
                ),
                remainder="passthrough",
                verbose_feature_names_out=False
            ).set_output(transform="pandas")
            
            X = self.preprocessor.fit_transform(X)
            
        else:
            X = self.preprocessor.transform(X)
            
        self.n_feats = X.shape[1]
        return X    
        
    
    def get_data_loader(self, X, y, pairs=None, shuffle=True):
        
        assert "time" in y.dtype.names and "event" in y.dtype.names, \
            "y must be a structured array with 'time' and 'event' fields."
        
        # Convert X to numpy if pandas
        if hasattr(X, 'values'):
            X = X.values
        if pairs is not None:
            if hasattr(pairs, 'values'):
                pairs = pairs.values

        if pairs is not None:
            dataset = torch.utils.data.TensorDataset(
                torch.FloatTensor(X), 
                torch.FloatTensor(pairs),
                torch.BoolTensor(y["event"]),
                torch.FloatTensor(y["time"].copy()),
            )
        else:
            dataset = torch.utils.data.TensorDataset(
                torch.FloatTensor(X), 
                torch.BoolTensor(y["event"]),
                torch.FloatTensor(y["time"].copy()),
            )
        loader = torch.utils.data.DataLoader(dataset, batch_size=self.batch_size, shuffle=shuffle)
        
        return loader
    
    def train_epoch(self, model, train_loader, optimizer):
        model.train()
        total_loss = 0

        for X, events, times in tqdm(train_loader, leave=False):

            X, events, times = X.to(self.device), events.to(self.device), times.to(self.device)

            y_pred = model(X).squeeze()
            
            surv_probs = torch.cumprod(y_pred, dim=1)
            
            loss =  bce_surv_loss(surv_probs, events, times, self.eval_times)

            optimizer.zero_grad()
            loss.backward()
            optimizer.step()

            total_loss += loss.item()
            
        return total_loss / len(train_loader)
    
    def test_epoch(self, model, test_loader):
        model.eval()
        total_loss = 0
        preds = []
        
        with torch.no_grad():
            for X, events, times in tqdm(test_loader, leave=False):

                X, events, times = X.to(self.device), events.to(self.device), times.to(self.device)

                y_pred = model(X).squeeze()
        
                preds.append(y_pred.detach())
            
                surv_probs = torch.cumprod(y_pred, dim=1)
        
                loss =  bce_surv_loss(surv_probs, events, times, self.eval_times)
                    
                total_loss += loss.item() 
        
        return total_loss / len(test_loader), torch.cat(preds)
    
    def fit_one_split(self, X_train, y_train, X_val, y_val):
        # If selected_feats is set, only use those features
        if hasattr(self, 'selected_feats'):
            X_train = X_train[self.selected_feats]
            X_val = X_val[self.selected_feats]
        
        # Get data loaders
        train_loader = self.get_data_loader(X_train, y_train)
        val_loader = self.get_data_loader(X_val, y_val, shuffle=False)
        
        model = SATransformerSingleSplit(
            n_feats=self.n_feats, 
            d_model=self.d_model, 
            n_eval_times=self.n_eval_times, 
            n_layers=self.n_layers, 
            num_heads=self.num_heads, 
            d_ff=self.d_ff, 
            drop_prob=self.drop_prob, 
            device=self.device
        ).to(self.device)
        
        
        optimizer = torch.optim.Adam(model.parameters(), lr=self.learning_rate)
        model.train_(
            train_epoch_fn=self.train_epoch,
            test_epoch_fn=self.test_epoch,
            train_loader=train_loader,
            test_loader=val_loader,
            optimizer=optimizer,
            n_epochs=self.max_epochs
        )
        
        return model
    
    def fit(self, X, y):
        
        # Preprocess features
        X = self.preprocess_data(X)
        
        self.feature_names_in_ = X.columns
        self.n_feats = X.shape[1]
        
                # Get evaluation times before train/val split
        quantiles = torch.quantile(
            torch.FloatTensor(y["time"].copy()),
            torch.linspace(0, 1, self.n_eval_times+2)
        )

        self.eval_times = quantiles[1:-1].to(self.device)
        
        # Remove duplicates in eval times
        if len(self.eval_times.unique()) < len(self.eval_times):
            self.eval_times = self.eval_times.unique()
            self.n_eval_times = len(self.eval_times)
        
        
        # Fit several models, one for each validation split
        self.models = []
        for i in range(self.n_val_splits):
            print("SPlIT", i)
        
            # Split the data into training and validation
            X_train, X_val, y_train, y_val = train_test_split(X, y, test_size=self.validation_size, random_state=10+i)
            
            # Fit to this split
            model = self.fit_one_split(X_train, y_train, X_val, y_val)
            model.feature_names_in_ = X_train.columns   
            
            self.models.append(model)
            
        return
    
    def predict(self, X_test):
        
        # Preprocess features
        X_test = self.preprocess_data(X_test)
        
        # Make placeholder y_test
        y_test = np.zeros(X_test.shape[0], dtype=[("event", "?"), ("time", "f8")])
        
        test_loader = self.get_data_loader(X_test, y_test, shuffle=False)
        
        test_preds = np.zeros((self.n_val_splits, X_test.shape[0], len(self.eval_times)))
        for i, model in enumerate(self.models):
            _, model_preds = self.test_epoch(model, test_loader)
            test_preds[i, ...] = model_preds.cpu().numpy()
            
        return np.mean(test_preds, axis=0)
    
    def get_calibration_data(self, X, y, eval_time, n_bins=20, method="quantile"):
        
        # First get cdf preds
        cdf_preds = 1 - np.cumprod(self.predict(X), axis=1)
        eval_index = np.searchsorted(self.eval_times.cpu().numpy(), eval_time)
        cdf_preds = cdf_preds[:, eval_index]
        
        
        if method == "quantile":
            # Get bins from quantiles of predictions
            quantiles = np.quantile(
                cdf_preds,
                np.linspace(0, 1, n_bins+2)
            )
        elif method == "uniform":
            # Get bins from uniform spacing
            quantiles = np.linspace(cdf_preds.min(), cdf_preds.max(), n_bins+2)
        else:
            raise ValueError("method must be 'quantile' or 'uniform'")
        
        predicted = []
        observed = []
        
        for bin_ in zip(quantiles[:-1], quantiles[1:]):
            
            if bin_[0] == bin_[1]:
                bin_data = cdf_preds[
                    cdf_preds == bin_[0]
                ]
                bin_y = y[
                    cdf_preds == bin_[0]
                ]
                
            else:
                bin_data = cdf_preds[
                    np.logical_and(
                        cdf_preds >= bin_[0],
                        cdf_preds < bin_[1]
                    )
                ]
                bin_y = y[
                    np.logical_and(
                        cdf_preds >= bin_[0],
                        cdf_preds < bin_[1]
                    )
                ]

            
            times, surv_prob = kaplan_meier_estimator(bin_y["event"], bin_y["time"])
            predicted.append(bin_data.mean())
            observed.append(
                1 - surv_prob[np.clip(
                    np.searchsorted(times, eval_time), 0, len(times)-1
                )]
            )
            
            
        return np.array(predicted), np.array(observed), quantiles